from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from enum import Enum
from typing import Dict
from typing import List
from uuid import UUID


class MetadataValueType(str, Enum):
    LOCATION = "LOCATION"  # string
    STRING = "STRING"  # string
    LINK = "LINK"  # string
    BOOLEAN = "BOOLEAN"  # bool
    NUMBER = "NUMBER"  # float
    DATE = "DATE"  # datetime


@dataclass(frozen=True)
class MetadataValue:
    value: str
    type_: MetadataValueType


class FileState(str, Enum):
    OK = "OK"
    CORRUPTED = "CORRUPTED"
    UPLOADING = "UPLOADING"
    ERROR = "ERROR"
    CONVERTING = "CONVERTING"
    CONVERSION_ERROR = "CONVERSION_ERROR"
    LOST = "LOST"
    FOUND = "FOUND"


@dataclass(frozen=True)
class Project:
    id: UUID
    name: str
    description: str
    created_at: datetime
    updated_at: datetime
    required_tags: List[str]


@dataclass(frozen=True)
class Mission:
    id: UUID
    name: str
    created_at: datetime
    updated_at: datetime
    project_id: UUID
    project_name: str
    metadata: Dict[str, MetadataValue] = field(default_factory=dict)
    number_of_files: int = 0
    size: int = 0


@dataclass(frozen=True)
class File:
    id: UUID
    name: str
    hash: str
    size: int
    type_: str
    date: datetime
    created_at: datetime
    updated_at: datetime
    mission_id: UUID
    mission_name: str
    project_id: UUID
    project_name: str
    categories: List[str] = field(default_factory=list)
    topics: List[str] = field(default_factory=list)
    state: FileState = FileState.OK


class RunStatus(str, Enum):
    QUEUED = "Queued"
    IN_PROGRESS = "In Progress"
    SUCCESS = "Success"
    FAILED = "Failed"
    CANCELLED = "Cancelled"


@dataclass(frozen=True)
class LogEntry:
    timestamp: datetime
    level: str
    message: str


@dataclass(frozen=True)
class Run:
    uuid: UUID
    state: str
    state_cause: str | None
    artifact_url: str | None
    created_at: datetime
    updated_at: datetime | None
    project_name: str
    mission_id: UUID
    mission_name: str
    template_id: UUID
    template_name: str
    logs: List[LogEntry] = field(default_factory=list)


@dataclass(frozen=True)
class ActionTemplate:
    uuid: UUID
    access_rights: int
    command: str
    cpu_cores: int
    cpu_memory_gb: int
    entrypoint: str
    gpu_memory_gb: int
    image_name: str
    max_runtime_minutes: int
    created_at: datetime
    name: str
    version: str


# this is the file state for the verify command
class FileVerificationStatus(str, Enum):
    UPLOADED = "uploaded"
    UPLOADING = "uploading"
    COMPUTING_HASH = "computing hash"
    MISSING = "missing"
    MISMATCHED_HASH = "hash mismatch"
    MISMATCHED_SIZE = "size mismatch"
    UNKNOWN = "unknown"



from __future__ import annotations

from pathlib import Path
from typing import List
from typing import Optional

import typer

import kleinkram.core
import kleinkram.utils
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.cli._file_validator import FileValidator
from kleinkram.cli._file_validator import _report_skipped_files
from kleinkram.config import get_shared_state
from kleinkram.errors import MissionNotFound
from kleinkram.utils import load_metadata
from kleinkram.utils import split_args

HELP = """\
Upload files to kleinkram.
"""

upload_typer = typer.Typer(
    name="upload",
    no_args_is_help=True,
    invoke_without_command=True,
    help=HELP,
)


def _build_mission_query(mission: str, project: Optional[str]) -> MissionQuery:
    """Constructs the MissionQuery object from CLI args."""
    mission_ids, mission_patterns = split_args([mission])
    project_ids, project_patterns = split_args([project] if project else [])

    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)
    return MissionQuery(
        ids=mission_ids,
        patterns=mission_patterns,
        project_query=project_query,
    )


def _handle_no_files_to_upload(original_count: int, uploaded_count: int) -> None:
    """Checks if any files are left to upload and exits if not."""
    if uploaded_count > 0:
        return

    if original_count > 0:
        typer.echo(
            typer.style("All paths were skipped. No files to upload.", fg=typer.colors.RED),
            err=True,
        )
    else:
        typer.echo(typer.style("No files provided to upload.", fg=typer.colors.RED), err=True)
    raise typer.Exit(code=1)


@upload_typer.callback()
def upload(
    files: List[str] = typer.Argument(help="files to upload"),
    project: Optional[str] = typer.Option(None, "--project", "-p", help="project id or name"),
    mission: str = typer.Option(..., "--mission", "-m", help="mission id or name"),
    create: bool = typer.Option(False, help="create mission if it does not exist"),
    metadata: Optional[str] = typer.Option(None, help="path to metadata file (json or yaml)"),
    fix_filenames: bool = typer.Option(
        False,
        help="fix filenames before upload, this does not change the filenames locally",
    ),
    skip: bool = typer.Option(
        False,
        "--skip",
        "-s",
        help="skip unsupported file types, badly named files, or directories instead of erroring",
    ),
    experimental_datatypes: bool = typer.Option(False, help="allow experimental datatypes (yaml, svo2, db3, tum)"),
    ignore_missing_tags: bool = typer.Option(False, help="ignore mission tags"),
) -> None:
    original_file_paths = [Path(file) for file in files]
    mission_query = _build_mission_query(mission, project)

    validator = FileValidator(
        skip=skip,
        experimental_datatypes=experimental_datatypes,
    )

    # This function will raise an error if skip=False and a file is invalid
    files_to_upload = validator.filter_files(original_file_paths)

    _report_skipped_files(validator.skipped_files)

    _handle_no_files_to_upload(original_count=len(original_file_paths), uploaded_count=len(files_to_upload))

    try:
        kleinkram.core.upload(
            client=AuthenticatedClient(),
            query=mission_query,
            file_paths=files_to_upload,
            create=create,
            metadata=load_metadata(Path(metadata)) if metadata else None,
            ignore_missing_metadata=ignore_missing_tags,
            verbose=get_shared_state().verbose,
        )
        typer.echo(
            typer.style(
                f"\nSuccessfully uploaded {len(files_to_upload)} file(s).",
                fg=typer.colors.GREEN,
            )
        )

    except MissionNotFound:
        if create:
            raise  # dont change the error message
        raise MissionNotFound("Mission not found. Use `--create` to create it.")



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import NewType
from typing import Tuple
from uuid import UUID

import dateutil.parser

from kleinkram.errors import ParsingError
from kleinkram.models import ActionTemplate
from kleinkram.models import File
from kleinkram.models import FileState
from kleinkram.models import LogEntry
from kleinkram.models import MetadataValue
from kleinkram.models import Mission
from kleinkram.models import Project
from kleinkram.models import Run

__all__ = [
    "_parse_project",
    "_parse_mission",
    "_parse_file",
]


ProjectObject = NewType("ProjectObject", Dict[str, Any])
MissionObject = NewType("MissionObject", Dict[str, Any])
FileObject = NewType("FileObject", Dict[str, Any])
RunObject = NewType("RunObject", Dict[str, Any])

MISSION = "mission"
PROJECT = "project"


class FileObjectKeys(str, Enum):
    UUID = "uuid"
    FILENAME = "filename"
    DATE = "date"  # at some point this will become a metadata
    CREATED_AT = "createdAt"
    UPDATED_AT = "updatedAt"
    STATE = "state"
    SIZE = "size"
    HASH = "hash"
    TYPE = "type"
    CATEGORIES = "categories"


class MissionObjectKeys(str, Enum):
    UUID = "uuid"
    NAME = "name"
    DESCRIPTION = "description"
    CREATED_AT = "createdAt"
    UPDATED_AT = "updatedAt"
    TAGS = "tags"
    FILESIZE = "size"
    FILECOUNT = "filesCount"


class ProjectObjectKeys(str, Enum):
    UUID = "uuid"
    NAME = "name"
    DESCRIPTION = "description"
    CREATED_AT = "createdAt"
    UPDATED_AT = "updatedAt"
    REQUIRED_TAGS = "requiredTags"


class RunObjectKeys(str, Enum):
    UUID = "uuid"
    STATE = "state"
    STATE_CAUSE = "stateCause"
    CREATED_AT = "createdAt"
    MISSION = "mission"
    TEMPLATE = "template"
    UPDATED_AT = "updatedAt"
    LOGS = "logs"
    ARTIFACT_URL = "artifactUrl"


class TemplateObjectKeys(str, Enum):
    UUID = "uuid"
    NAME = "name"
    ACCESS_RIGHTS = "accessRights"
    COMMAND = "command"
    CPU_CORES = "cpuCores"
    CPU_MEMORY_GB = "cpuMemory"
    ENTRYPOINT = "entrypoint"
    GPU_MEMORY_GB = "gpuMemory"
    IMAGE_NAME = "imageName"
    MAX_RUNTIME_MINUTES = "maxRuntime"
    CREATED_AT = "createdAt"
    VERSION = "version"


class LogEntryObjectKeys(str, Enum):
    TIMESTAMP = "timestamp"
    LEVEL = "type"
    MESSAGE = "message"


def _get_nested_info(data, key: Literal["mission", "project"]) -> Tuple[UUID, str]:
    nested_data = data[key]
    return (
        UUID(nested_data[ProjectObjectKeys.UUID], version=4),
        nested_data[ProjectObjectKeys.NAME],
    )


def _parse_datetime(date: str) -> datetime:
    try:
        return dateutil.parser.isoparse(date)
    except ValueError as e:
        raise ParsingError(f"error parsing date: {date}") from e


def _parse_file_state(state: str) -> FileState:
    try:
        return FileState(state)
    except ValueError as e:
        raise ParsingError(f"error parsing file state: {state}") from e


def _parse_metadata(tags: List[Dict]) -> Dict[str, MetadataValue]:
    result = {}
    try:
        for tag in tags:
            entry = {tag.get("name"): MetadataValue(tag.get("valueAsString"), tag.get("datatype"))}
            result.update(entry)
        return result
    except ValueError as e:
        raise ParsingError(f"error parsing metadata: {e}") from e


def _parse_required_tags(tags: List[Dict]) -> list[str]:
    return list(_parse_metadata(tags).keys())


def _parse_project(project_object: ProjectObject) -> Project:
    try:
        id_ = UUID(project_object[ProjectObjectKeys.UUID], version=4)
        name = project_object[ProjectObjectKeys.NAME]
        description = project_object[ProjectObjectKeys.DESCRIPTION]
        created_at = _parse_datetime(project_object[ProjectObjectKeys.CREATED_AT])
        updated_at = _parse_datetime(project_object[ProjectObjectKeys.UPDATED_AT])
        required_tags = _parse_required_tags(project_object[ProjectObjectKeys.REQUIRED_TAGS])
    except Exception as e:
        raise ParsingError(f"error parsing project: {project_object}") from e
    return Project(
        id=id_,
        name=name,
        description=description,
        created_at=created_at,
        updated_at=updated_at,
        required_tags=required_tags,
    )


def _parse_mission(mission: MissionObject) -> Mission:
    try:
        id_ = UUID(mission[MissionObjectKeys.UUID], version=4)
        name = mission[MissionObjectKeys.NAME]
        created_at = _parse_datetime(mission[MissionObjectKeys.CREATED_AT])
        updated_at = _parse_datetime(mission[MissionObjectKeys.UPDATED_AT])
        metadata = _parse_metadata(mission[MissionObjectKeys.TAGS])
        file_count = mission[MissionObjectKeys.FILECOUNT]
        filesize = mission[MissionObjectKeys.FILESIZE]

        project_id, project_name = _get_nested_info(mission, PROJECT)

        parsed = Mission(
            id=id_,
            name=name,
            created_at=created_at,
            updated_at=updated_at,
            metadata=metadata,
            project_id=project_id,
            project_name=project_name,
            number_of_files=file_count,
            size=filesize,
        )
    except Exception as e:
        raise ParsingError(f"error parsing mission: {mission}") from e
    return parsed


def _parse_file(file: FileObject) -> File:
    try:
        name = file[FileObjectKeys.FILENAME]
        id_ = UUID(file[FileObjectKeys.UUID], version=4)
        fsize = file[FileObjectKeys.SIZE]
        fhash = file[FileObjectKeys.HASH]
        ftype = file[FileObjectKeys.TYPE].split(".")[-1]
        fdate = file[FileObjectKeys.DATE]
        created_at = _parse_datetime(file[FileObjectKeys.CREATED_AT])
        updated_at = _parse_datetime(file[FileObjectKeys.UPDATED_AT])
        state = _parse_file_state(file[FileObjectKeys.STATE])
        categories = file[FileObjectKeys.CATEGORIES]

        mission_id, mission_name = _get_nested_info(file, MISSION)
        project_id, project_name = _get_nested_info(file[MISSION], PROJECT)

        parsed = File(
            id=id_,
            name=name,
            hash=fhash,
            size=fsize,
            type_=ftype,
            date=fdate,
            categories=categories,
            state=state,
            created_at=created_at,
            updated_at=updated_at,
            mission_id=mission_id,
            mission_name=mission_name,
            project_id=project_id,
            project_name=project_name,
        )
    except Exception as e:
        raise ParsingError(f"error parsing file: {file}") from e
    return parsed


def _parse_action_template(run_object: RunObject) -> ActionTemplate:
    try:
        uuid_ = UUID(run_object[TemplateObjectKeys.UUID], version=4)
        access_rights = run_object[TemplateObjectKeys.ACCESS_RIGHTS]
        command = run_object[TemplateObjectKeys.COMMAND]
        cpu_cores = run_object[TemplateObjectKeys.CPU_CORES]
        cpu_memory_gb = run_object[TemplateObjectKeys.CPU_MEMORY_GB]
        entrypoint = run_object[TemplateObjectKeys.ENTRYPOINT]
        gpu_memory_gb = run_object[TemplateObjectKeys.GPU_MEMORY_GB]
        image_name = run_object[TemplateObjectKeys.IMAGE_NAME]
        max_runtime_minutes = run_object[TemplateObjectKeys.MAX_RUNTIME_MINUTES]
        created_at = _parse_datetime(run_object[TemplateObjectKeys.CREATED_AT])
        name = run_object[TemplateObjectKeys.NAME]
        version = run_object[TemplateObjectKeys.VERSION]

    except Exception as e:
        raise ParsingError(f"error parsing action template: {run_object}") from e

    return ActionTemplate(
        uuid=uuid_,
        access_rights=access_rights,
        command=command,
        cpu_cores=cpu_cores,
        cpu_memory_gb=cpu_memory_gb,
        entrypoint=entrypoint,
        gpu_memory_gb=gpu_memory_gb,
        image_name=image_name,
        max_runtime_minutes=max_runtime_minutes,
        created_at=created_at,
        name=name,
        version=version,
    )


def _parse_run(run_object: RunObject) -> Run:
    try:
        uuid_ = UUID(run_object[RunObjectKeys.UUID], version=4)
        state = run_object[RunObjectKeys.STATE]
        state_cause = run_object[RunObjectKeys.STATE_CAUSE]
        artifact_url = run_object.get(RunObjectKeys.ARTIFACT_URL)
        created_at = _parse_datetime(run_object[RunObjectKeys.CREATED_AT])
        updated_at = (
            _parse_datetime(run_object[RunObjectKeys.UPDATED_AT]) if run_object.get(RunObjectKeys.UPDATED_AT) else None
        )

        mission_dict = run_object[RunObjectKeys.MISSION]
        mission_id = UUID(mission_dict[MissionObjectKeys.UUID], version=4)
        mission_name = mission_dict[MissionObjectKeys.NAME]

        project_dict = mission_dict[PROJECT]
        project_name = project_dict[ProjectObjectKeys.NAME]

        template_dict = run_object[RunObjectKeys.TEMPLATE]
        template_id = UUID(template_dict[TemplateObjectKeys.UUID], version=4)
        template_name = template_dict[TemplateObjectKeys.NAME]
        logs = []
        for log_entry in run_object.get(RunObjectKeys.LOGS, []):
            log_timestamp = _parse_datetime(log_entry[LogEntryObjectKeys.TIMESTAMP])
            log_level = log_entry[LogEntryObjectKeys.LEVEL]
            log_message = log_entry[LogEntryObjectKeys.MESSAGE]
            logs.append(
                LogEntry(
                    timestamp=log_timestamp,
                    level=log_level,
                    message=log_message,
                )
            )

    except Exception as e:
        raise ParsingError(f"error parsing run: {run_object}") from e

    return Run(
        uuid=uuid_,
        state=state,
        state_cause=state_cause,
        artifact_url=artifact_url,
        created_at=created_at,
        updated_at=updated_at,
        mission_id=mission_id,
        mission_name=mission_name,
        project_name=project_name,
        template_id=template_id,
        template_name=template_name,
        logs=logs,
    )



from __future__ import annotations

import urllib.parse
import webbrowser
from getpass import getpass
from http.server import BaseHTTPRequestHandler
from http.server import HTTPServer
from typing import Optional

from kleinkram.config import CONFIG_PATH
from kleinkram.config import Credentials
from kleinkram.config import get_config
from kleinkram.config import save_config

CLI_CALLBACK_ENDPOINT = "/cli/callback"
OAUTH_SLUG = "/auth/"


def _has_browser() -> bool:
    try:
        webbrowser.get()
        return True
    except webbrowser.Error:
        return False


def _headless_auth(*, url: str) -> None:

    print(f"please open the following URL manually to authenticate: {url}")
    print("enter the authentication token provided after logging in:")
    auth_token = getpass("authentication token: ")
    refresh_token = getpass("refresh token: ")

    if auth_token and refresh_token:
        config = get_config()
        config.credentials = Credentials(auth_token=auth_token, refresh_token=refresh_token)
        save_config(config)
        print(f"Authentication complete. Tokens saved to {CONFIG_PATH}.")
    else:
        raise ValueError("Please provided tokens.")


class OAuthCallbackHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        if self.path.startswith(CLI_CALLBACK_ENDPOINT):
            query = urllib.parse.urlparse(self.path).query
            params = urllib.parse.parse_qs(query)

            try:
                creds = Credentials(
                    auth_token=params.get("authtoken")[0],  # type: ignore
                    refresh_token=params.get("refreshtoken")[0],  # type: ignore
                )
                config = get_config()
                config.credentials = creds
                save_config(config)
            except Exception:
                raise RuntimeError("Failed to fetch authentication tokens.")

            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"Authentication successful. You can close this window.")
        else:
            raise RuntimeError("Invalid path")

    def log_message(self, *args, **kwargs):
        _ = args, kwargs
        pass  # suppress logging


def _browser_auth(*, url: str) -> None:
    webbrowser.open(url)

    server = HTTPServer(("", 8000), OAuthCallbackHandler)
    server.handle_request()

    print(f"Authentication complete. Tokens saved to {CONFIG_PATH}.")


def _direct_oauth_auth(*, endpoint: str, provider: str, user: str) -> None:
    """
    Directly authenticate with fake OAuth by programmatically following the OAuth flow.
    This bypasses the browser entirely for automated testing.
    """
    import requests

    print(f"Authenticating as user {user} with {provider}...")

    try:
        # Step 1: Get the authorization code from fake OAuth
        # The fake OAuth server will auto-redirect when user parameter is provided
        fake_oauth_url = "http://localhost:8004/oauth/authorize"
        callback_url = f"{endpoint}/auth/{provider}/callback"

        params = {
            "client_id": "some-random-string-it-does-not-matter",
            "redirect_uri": callback_url,
            "response_type": "code",
            "state": "cli-direct",
            "user": user,
        }

        # Make request to fake OAuth - it will redirect with the auth code
        response = requests.get(fake_oauth_url, params=params, allow_redirects=False)

        if response.status_code not in [301, 302, 303, 307, 308]:
            raise RuntimeError(f"Expected redirect from OAuth provider, got {response.status_code}")

        # Extract the redirect location
        location = response.headers.get("Location")
        if not location:
            raise RuntimeError("No redirect location from OAuth provider")

        # Parse the callback URL to extract the auth code
        parsed = urllib.parse.urlparse(location)
        query_params = urllib.parse.parse_qs(parsed.query)

        if "code" not in query_params:
            raise RuntimeError(f"No authorization code in redirect: {location}")

        auth_code = query_params["code"][0]
        state = query_params.get("state", [None])[0]

        print("Received authorization code, exchanging for tokens...")

        # Step 2: Exchange the code for tokens by calling the backend callback
        # Use a session to preserve cookies
        session = requests.Session()
        callback_params = {"code": auth_code}
        if state:
            callback_params["state"] = state

        callback_response = session.get(callback_url, params=callback_params, allow_redirects=False)

        # The backend should set cookies and redirect
        if callback_response.status_code not in [301, 302, 303, 307, 308]:
            raise RuntimeError(f"Expected redirect from callback, got {callback_response.status_code}")

        # Extract tokens from cookies
        auth_token = session.cookies.get("authtoken")
        refresh_token = session.cookies.get("refreshtoken")

        if not auth_token or not refresh_token:
            raise RuntimeError("Failed to get tokens from callback response")

        # Save tokens
        config = get_config()
        config.credentials = Credentials(auth_token=auth_token, refresh_token=refresh_token)
        save_config(config)
        print(f"Authentication complete. Tokens saved to {CONFIG_PATH}.")

    except requests.RequestException as e:
        raise RuntimeError(f"OAuth flow failed: {e}")


def login_flow(
    *,
    oAuthProvider: str,
    key: Optional[str] = None,
    headless: bool = False,
    user: Optional[str] = None,
) -> None:
    config = get_config()
    # use cli key login
    if key is not None:
        config.credentials = Credentials(api_key=key)
        save_config(config)
        return

    # If user parameter is provided with fake-oauth, use direct OAuth flow
    if user is not None and oAuthProvider == "fake-oauth":
        _direct_oauth_auth(endpoint=config.endpoint.api, provider=oAuthProvider, user=user)
        return

    # Build OAuth URL with state parameter
    oauth_url = f"{config.endpoint.api}{OAUTH_SLUG}{oAuthProvider}?state=cli"

    # Add user parameter if provided (for fake-oauth auto-login)
    if user is not None:
        oauth_url += f"&user={user}"

    is_port_available = True
    try:
        server = HTTPServer(("", 8000), OAuthCallbackHandler)
        server.server_close()
    except OSError:
        is_port_available = False

    if not is_port_available:
        print("Warning: Port 8000 is not available. Falling back to headless authentication.\n\n")

    if not headless and _has_browser() and is_port_available:
        _browser_auth(url=oauth_url)
    else:
        _headless_auth(url=f"{oauth_url}-no-redirect")



from __future__ import annotations

import os
import re
import sys
import tarfile
import time
from typing import List
from typing import Optional

import requests
import typer

import kleinkram.api.routes
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import RunQuery
from kleinkram.config import get_shared_state
from kleinkram.models import LogEntry
from kleinkram.models import Run
from kleinkram.printing import print_run_info
from kleinkram.printing import print_run_logs
from kleinkram.printing import print_runs_table
from kleinkram.utils import split_args

HELP = """\
Manage and inspect action runs.

You can list action runs, get detailed information about specific runs, stream their logs,
cancel runs in progress, and retry failed runs.
"""

run_typer = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    help=HELP,
)

LIST_HELP = "List action runs. Optionally filter by mission or project."
INFO_HELP = "Get detailed information about a specific action run."
LOGS_HELP = "Stream the logs for a specific action run."
CANCEL_HELP = "Cancel an action run that is in progress."
RETRY_HELP = "Retry a failed action run."
DOWNLOAD_HELP = "Download artifacts for a specific action run."


@run_typer.command(help=LIST_HELP, name="list")
def list_runs(
    mission: Optional[str] = typer.Option(None, "--mission", "-m", help="Mission ID or name to filter by."),
    project: Optional[str] = typer.Option(None, "--project", "-p", help="Project ID or name to filter by."),
) -> None:
    """
    List action runs.
    """
    client = AuthenticatedClient()

    mission_ids, mission_patterns = split_args([mission] if mission else [])
    project_ids, project_patterns = split_args([project] if project else [])

    query = RunQuery(
        mission_ids=mission_ids,
        mission_patterns=mission_patterns,
        project_ids=project_ids,
        project_patterns=project_patterns,
    )

    runs = list(kleinkram.api.routes.get_runs(client, query=query))
    print_runs_table(runs, pprint=get_shared_state().verbose)


@run_typer.command(name="info", help=INFO_HELP)
def get_info(run_id: str = typer.Argument(..., help="The ID of the run to get information for.")) -> None:
    """
    Get detailed information for a single run.
    """
    client = AuthenticatedClient()
    run: Run = kleinkram.api.routes.get_run(client, run_id=run_id)
    print_run_info(run, pprint=get_shared_state().verbose)


@run_typer.command(help=LOGS_HELP)
def logs(
    run_id: str = typer.Argument(..., help="The ID of the run to fetch logs for."),
    follow: bool = typer.Option(False, "--follow", "-f", help="Follow the log output in real-time."),
) -> None:
    """
    Fetch and display logs for a specific run.
    """
    client = AuthenticatedClient()

    if follow:
        typer.echo(f"Watching logs for run {run_id}. Press Ctrl+C to stop.")
        try:

            # TODO: fine for now, but ideally we would have a streaming endpoint
            # currently there is no following, thus we just poll every 2 seconds
            # from the get_run endpoint
            last_log_index = 0
            while True:
                run: Run = kleinkram.api.routes.get_run(client, run_id=run_id)
                log_entries: List[LogEntry] = run.logs
                new_log_entries = log_entries[last_log_index:]
                if new_log_entries:
                    print_run_logs(new_log_entries, pprint=get_shared_state().verbose)
                    last_log_index += len(new_log_entries)

                time.sleep(2)

        except KeyboardInterrupt:
            typer.echo("Stopped following logs.")
            sys.exit(0)
    else:
        log_entries = kleinkram.api.routes.get_run(client, run_id=run_id).logs
        print_run_logs(log_entries, pprint=get_shared_state().verbose)


def _get_filename_from_cd(cd: str) -> Optional[str]:
    """Extract filename from Content-Disposition header."""
    if not cd:
        return None
    fname = re.findall("filename=(.+)", cd)
    if len(fname) == 0:
        return None
    return fname[0].strip().strip('"')


@run_typer.command(name="download", help=DOWNLOAD_HELP)
def download_artifacts(
    run_id: str = typer.Argument(..., help="The ID of the run to download artifacts for."),
    output: Optional[str] = typer.Option(None, "--output", "-o", help="Path or filename to save the artifacts to."),
    extract: bool = typer.Option(
        False,
        "--extract",
        "-x",
        help="Automatically extract the archive after downloading.",
    ),
) -> None:
    """
    Download the artifacts (.tar.gz) for a finished run.
    """
    client = AuthenticatedClient()

    # Fetch Run Details
    try:
        run: Run = kleinkram.api.routes.get_run(client, run_id=run_id)
    except Exception as e:
        typer.secho(f"Failed to fetch run details: {e}", fg=typer.colors.RED)
        raise typer.Exit(1)

    if not run.artifact_url:
        typer.secho(
            f"No artifacts found for run {run_id}. The run might not be finished or artifacts expired.",
            fg=typer.colors.YELLOW,
        )
        raise typer.Exit(1)

    typer.echo(f"Downloading artifacts for run {run_id}...")

    # Stream Download
    try:
        with requests.get(run.artifact_url, stream=True) as r:
            r.raise_for_status()

            # Determine Filename
            filename = output
            if not filename:
                filename = _get_filename_from_cd(r.headers.get("content-disposition"))

            if not filename:
                filename = f"{run_id}.tar.gz"

            # If output is a directory, join with filename
            if output and os.path.isdir(output):
                filename = os.path.join(
                    output,
                    _get_filename_from_cd(r.headers.get("content-disposition")) or f"{run_id}.tar.gz",
                )

            total_length = int(r.headers.get("content-length", 0))

            # Write to file with Progress Bar
            with open(filename, "wb") as f:
                with typer.progressbar(length=total_length, label=f"Saving to {filename}") as progress:
                    for chunk in r.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
                            progress.update(len(chunk))

            typer.secho(f"\nSuccessfully downloaded to {filename}", fg=typer.colors.GREEN)

            # Extraction Logic
            if extract:
                try:
                    # Determine extraction directory (based on filename without extension)
                    # e.g., "downloads/my-run.tar" -> "downloads/my-run"
                    base_name = os.path.basename(filename)
                    folder_name = base_name.split(".")[0]

                    # Get the parent directory of the downloaded file
                    parent_dir = os.path.dirname(os.path.abspath(filename))
                    extract_path = os.path.join(parent_dir, folder_name)

                    typer.echo(f"Extracting to: {extract_path}...")

                    with tarfile.open(filename, "r:gz") as tar:

                        # Safety check: filter_data prevents extraction outside target dir (CVE-2007-4559)
                        # Available in Python 3.12+, for older python use generic extractall
                        if hasattr(tarfile, "data_filter"):
                            tar.extractall(path=extract_path, filter="data")
                        else:
                            tar.extractall(path=extract_path)

                    typer.secho("Successfully extracted.", fg=typer.colors.GREEN)

                except tarfile.TarError as e:
                    typer.secho(f"Failed to extract archive: {e}", fg=typer.colors.RED)

    except requests.exceptions.RequestException as e:
        typer.secho(f"Error downloading file: {e}", fg=typer.colors.RED)
        raise typer.Exit(1)



from __future__ import annotations

import os
import secrets
import shutil
import time
from pathlib import Path

import pytest
from rich.console import Console
from rich.text import Text

from kleinkram.api.routes import _get_api_version

VERBOSE = True

CLI = "klein"
PROJECT_NAME = "automated-testing"
DATA_PATH = Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def api():
    try:
        _get_api_version()
        return True
    except Exception:
        print("API is not available")
        return False


def run_cmd(command, *, verbose=VERBOSE):
    msg = ("\n", "#" * 50, "\n\n", "running command:", Text(command, style="bold"))
    Console().print(*msg)

    if not verbose:
        command += ">/dev/null 2>&1"
    ret = os.system(command)
    Console().print("got return code:", ret, style="bold red")
    return ret


@pytest.mark.slow
def test_upload_verify_update_download_mission(project, tmp_path, api):
    assert api

    file_names = list(DATA_PATH.glob("*.bag"))

    mission_name = secrets.token_hex(8)
    upload = f"{CLI} upload -p {project.name} -m {mission_name} --create {DATA_PATH.absolute()}/*.bag"
    verify = f"{CLI} verify -p {project.name} -m {mission_name} {DATA_PATH.absolute()}/*.bag"
    # update = f"{CLI} mission update -p {project.name} -m {mission_name} --metadata {DATA_PATH.absolute()}/metadata.yaml"
    download = f"{CLI} download -p {project.name} -m {mission_name} --dest {tmp_path.absolute()}"
    delete_file = f"{CLI} file delete -p {project.name} -m {mission_name} -f {file_names[0].name} -y"

    assert run_cmd(upload) == 0
    assert run_cmd(verify) == 0
    # assert run_cmd(update) == 0
    assert run_cmd(download) == 0

    assert run_cmd(delete_file) == 0


@pytest.mark.slow
def test_list_files(project, mission, api):
    assert api
    assert run_cmd(f"{CLI} list files -p {project.name}") == 0
    assert run_cmd(f"{CLI} list files -p {project.name} -m {mission.name}") == 0
    assert run_cmd(f"{CLI} list files") == 0
    assert run_cmd(f"{CLI} list files -p {project.name}") == 0
    assert run_cmd(f'{CLI} list files -p "*" -m "*" "*"') == 0


@pytest.mark.slow
def test_list_missions(api, project, mission):
    assert api

    assert run_cmd(f"{CLI} list missions -p {project.name} {mission.name}") == 0
    assert run_cmd(f"{CLI} list missions -p {project.name} {secrets.token_hex(8)}") == 0
    assert run_cmd(f"{CLI} list missions -p {project.name} {mission.id}") == 0
    assert run_cmd(f"{CLI} list missions {secrets.token_hex(8)}") == 0
    assert run_cmd(f"{CLI} list missions {mission.id}") == 0
    assert run_cmd(f"{CLI} list missions {mission.name}") == 0

    assert run_cmd(f"{CLI} list missions -p {project.name}") == 0
    assert run_cmd(f"{CLI} list missions -p {project.id}") == 0
    assert run_cmd(f"{CLI} list missions -p {secrets.token_hex(8)}") == 0
    assert run_cmd(f"{CLI} list missions") == 0
    assert run_cmd(f'{CLI} list missions -p "*" "*"') == 0


@pytest.mark.slow
def test_list_projects(api, project):
    assert api
    assert run_cmd(f"{CLI} list projects") == 0
    assert run_cmd(f"{CLI} list projects {project.name}") == 0
    assert run_cmd(f"{CLI} list projects {secrets.token_hex(8)}") == 0
    assert run_cmd(f"{CLI} list projects {project.id}") == 0
    assert run_cmd(f'{CLI} list projects "*"') == 0



from __future__ import annotations

import time
from typing import Optional
from uuid import UUID

import httpx
import typer

import kleinkram.api.routes
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.config import get_shared_state
from kleinkram.printing import print_action_templates_table
from kleinkram.printing import print_run_info
from kleinkram.utils import is_valid_uuid4
from kleinkram.utils import split_args

HELP = """\
Launch kleinkram actions from predefined templates.

You can list available action templates, launch new actions on specific missions, and optionally
follow their logs in real-time.
"""

action_typer = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    help=HELP,
)

LIST_HELP = "Lists action templates (definitions). To list individual runs, use `klein run list`."
GET_HELP = "Get details for a specific action template."
RUN_HELP = "Launch a new action from a template."


@action_typer.command(help=LIST_HELP, name="list")
def list_actions() -> None:
    client = AuthenticatedClient()
    templates = list(kleinkram.api.routes.get_action_templates(client))

    if not templates:
        typer.echo("No action templates found.")
        return

    print_action_templates_table(templates, pprint=get_shared_state().verbose)


@action_typer.command(help=RUN_HELP)
def run(
    template_name: str = typer.Argument(..., help="Name or ID of the template to run."),
    mission: str = typer.Option(..., "--mission", "-m", help="Mission ID or name to run the action on."),
    project: Optional[str] = typer.Option(None, "--project", "-p", help="Project ID or name (to scope mission)."),
    follow: bool = typer.Option(False, "--follow", "-f", help="Follow the logs of the action run."),
) -> None:
    """
    Submits an action to run on a specific mission and optionally follows its logs.
    """
    client = AuthenticatedClient()
    pprint = get_shared_state().verbose

    try:
        project_ids, project_patterns = split_args([project] if project else [])
        project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)

        mission_ids, mission_patterns = split_args([mission])
        mission_query = MissionQuery(
            ids=mission_ids,
            patterns=mission_patterns,
            project_query=project_query,
        )
        mission_obj = kleinkram.api.routes.get_mission(client, mission_query)
        mission_uuid = mission_obj.id
    except kleinkram.errors.MissionNotFound:
        typer.secho(f"Error: Mission '{mission}' not found.", fg=typer.colors.RED)
        raise typer.Exit(code=1)
    except kleinkram.errors.InvalidMissionQuery:
        typer.secho(
            "Error: Mission query is ambiguous. Try specifying a project with -p.",
            fg=typer.colors.RED,
        )
        raise typer.Exit(code=1)
    except Exception as e:
        typer.secho(f"Error resolving mission: {e}", fg=typer.colors.RED)
        raise typer.Exit(code=1)

    # 2. Resolve Template to UUID
    try:
        if is_valid_uuid4(template_name):
            template_uuid = UUID(template_name)
        else:
            templates = kleinkram.api.routes.get_action_templates(client)
            found_template = next((t for t in templates if t.name == template_name), None)

            if not found_template:
                typer.secho(
                    f"Error: Action template '{template_name}' not found.",
                    fg=typer.colors.RED,
                )
                raise typer.Exit(code=1)
            template_uuid = found_template.uuid
    except Exception as e:
        typer.secho(f"Error resolving template: {e}", fg=typer.colors.RED)
        raise typer.Exit(code=1)

    try:
        action_uuid_str = kleinkram.api.routes.submit_action(client, mission_uuid, template_uuid)
        typer.secho(f"Action submitted. Run ID: {action_uuid_str}", fg=typer.colors.GREEN)

    except httpx.HTTPStatusError as e:
        typer.secho(f"Error submitting action: {e.response.text}", fg=typer.colors.RED)
        raise typer.Exit(code=1)
    except (KeyError, Exception) as e:
        typer.secho(f"An unexpected error occurred: {e}", fg=typer.colors.RED)
        raise typer.Exit(code=1)

    if follow:
        exit_code = kleinkram.printing.follow_run_logs(client, action_uuid_str)
        if exit_code != 0:
            raise typer.Exit(code=exit_code)

    elif pprint:
        # Not following, but in verbose mode. Show run info.
        try:
            time.sleep(0.5)  # Give API a moment
            run_details = kleinkram.api.routes.get_run(client, action_uuid_str)
            kleinkram.printing.print_run_info(run_details, pprint=True)
        except Exception:
            # Non-critical, we already printed the ID.
            pass



"""
this file contains a global config and a global state object

to get the config use `get_config()`
"""

from __future__ import annotations

import json
import logging
import os
import tempfile
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from pathlib import Path
from typing import Any
from typing import Dict
from typing import NamedTuple
from typing import Optional

from rich.table import Table
from rich.text import Text

from kleinkram._version import __local__
from kleinkram._version import __version__
from kleinkram.utils import format_traceback

logger = logging.getLogger(__name__)

CONFIG_PATH = Path().home() / ".kleinkram.json"
MAX_TABLE_SIZE = 256


class Environment(Enum):
    LOCAL = "local"
    DEV = "dev"
    PROD = "prod"


class Endpoint(NamedTuple):
    name: str
    api: str
    s3: str


class Credentials(NamedTuple):
    auth_token: Optional[str] = None
    refresh_token: Optional[str] = None
    api_key: Optional[str] = None


DEFAULT_LOCAL_API = "http://localhost:3000"
DEFAULT_LOCAL_S3 = "http://localhost:9000"

DEFAULT_DEV_API = "https://api.datasets.dev.leggedrobotics.com"
DEFAULT_DEV_S3 = "https://s3.datasets.dev.leggedrobotics.com"

DEFAULT_PROD_API = "https://api.datasets.leggedrobotics.com"
DEFAULT_PROD_S3 = "https://s3.datasets.leggedrobotics.com"


DEFAULT_ENDPOINTS = {
    "local": Endpoint("local", DEFAULT_LOCAL_API, DEFAULT_LOCAL_S3),
    "dev": Endpoint("dev", DEFAULT_DEV_API, DEFAULT_DEV_S3),
    "prod": Endpoint("prod", DEFAULT_PROD_API, DEFAULT_PROD_S3),
}


def get_env() -> Environment:
    if __local__:
        return Environment.LOCAL
    if "dev" in __version__:
        return Environment.DEV
    return Environment.PROD


ACTION_API_KEY = "KLEINKRAM_API_KEY"
ACTION_API = "KLEINKRAM_API_ENDPOINT"
ACTION_S3 = "KLEINKRAM_S3_ENDPOINT"


def _get_endpoint_from_action_env_vars() -> Optional[Endpoint]:
    api = os.getenv(ACTION_API)
    s3 = os.getenv(ACTION_S3)
    if api is None or s3 is None:
        return None
    return Endpoint("action", api, s3)


def _get_api_key_from_action_env_vars() -> Optional[str]:
    return os.getenv(ACTION_API_KEY)


def _get_default_selected_endpoint() -> Endpoint:
    env_endpoint = _get_endpoint_from_action_env_vars()
    if env_endpoint is not None:
        return env_endpoint
    return DEFAULT_ENDPOINTS[get_env().value]


def _get_default_endpoints() -> Dict[str, Endpoint]:
    env_endpoint = _get_endpoint_from_action_env_vars()

    default_endpoints = DEFAULT_ENDPOINTS.copy()
    if env_endpoint is not None:
        default_endpoints["action"] = env_endpoint
    return default_endpoints


def _get_default_credentials() -> Dict[str, Credentials]:
    endpoint = _get_default_selected_endpoint()

    api_key = _get_api_key_from_action_env_vars()
    if api_key is not None:
        return {endpoint.name: Credentials(api_key=api_key)}
    return {}


@dataclass
class Config:
    version: str = __version__
    selected_endpoint: str = field(default_factory=lambda: _get_default_selected_endpoint().name)
    endpoints: Dict[str, Endpoint] = field(default_factory=_get_default_endpoints)
    endpoint_credentials: Dict[str, Credentials] = field(default_factory=_get_default_credentials)

    @property
    def endpoint(self) -> Endpoint:
        return self.endpoints[self.selected_endpoint]

    @endpoint.setter
    def endpoint(self, value: Endpoint) -> None:
        self.endpoints[self.selected_endpoint] = value

    @property
    def credentials(self) -> Optional[Credentials]:
        return self.endpoint_credentials.get(self.selected_endpoint)

    @credentials.setter
    def credentials(self, value: Credentials) -> None:
        self.endpoint_credentials[self.selected_endpoint] = value


def _config_to_dict(config: Config) -> Dict[str, Any]:
    return {
        "version": config.version,
        "endpoints": {key: value._asdict() for key, value in config.endpoints.items()},
        "endpoint_credentials": {key: value._asdict() for key, value in config.endpoint_credentials.items()},
        "selected_endpoint": config.endpoint.name,
    }


def _config_from_dict(dct: Dict[str, Any]) -> Config:
    return Config(
        dct["version"],
        dct["selected_endpoint"],
        {key: Endpoint(**value) for key, value in dct["endpoints"].items()},
        {key: Credentials(**value) for key, value in dct["endpoint_credentials"].items()},
    )


def _safe_config_write(config: Config, path: Path, tmp_dir: Optional[Path] = None) -> None:
    fd, temp_path = tempfile.mkstemp(dir=tmp_dir)
    with os.fdopen(fd, "w") as f:
        json.dump(_config_to_dict(config), f)
    os.replace(temp_path, path)


def _unsafe_config_write(config: Config, path: Path) -> None:
    with open(path, "w") as f:
        json.dump(_config_to_dict(config), f)


def save_config(config: Config, path: Path = CONFIG_PATH) -> None:
    try:
        _safe_config_write(config, path)
    except Exception as e:
        logger.debug(f"failed to safe write config {format_traceback(e)}")
        _unsafe_config_write(config, path)


def _load_config_if_compatible(path: Path) -> Optional[Config]:
    if not path.exists():
        return None
    with open(path, "r") as f:
        try:
            dct = json.load(f)

            migrated = False
            for key, value in dct.get("endpoints", {}).items():
                if "s3" in value and "minio.datasets." in value["s3"]:
                    value["s3"] = value["s3"].replace("minio.datasets.", "s3.datasets.")
                    migrated = True

            config = _config_from_dict(dct)
            if migrated:
                save_config(config, path)
            return config
        except Exception:
            return None


def _load_config(*, path: Path = CONFIG_PATH) -> Config:
    config = _load_config_if_compatible(path)
    if config is None:
        return Config()
    return config


LOADED_CONFIGS: Dict[Path, Config] = {}


def get_config(path: Path = CONFIG_PATH) -> Config:
    if path not in LOADED_CONFIGS:
        LOADED_CONFIGS[path] = _load_config(path=path)
    return LOADED_CONFIGS[path]


def select_endpoint(config: Config, name: str, path: Path = CONFIG_PATH) -> None:
    if name not in config.endpoints:
        raise ValueError(f"Endpoint {name} not found.")
    config.selected_endpoint = name
    save_config(config, path)


def add_endpoint(config: Config, endpoint: Endpoint, path: Path = CONFIG_PATH) -> None:
    config.endpoints[endpoint.name] = endpoint
    config.selected_endpoint = endpoint.name
    save_config(config, path)


def check_config_compatibility(path: Path = CONFIG_PATH) -> bool:
    """\
    returns `False` if config file exists but is not compatible with the current version

    TODO: add more sophisticated version checking
    """
    if not path.exists():
        return True
    config = _load_config_if_compatible(path)
    return config is not None


def endpoint_table(config: Config) -> Table:
    table = Table(title="Available Endpoints")
    table.add_column("Name", style="cyan")
    table.add_column("API", style="cyan")
    table.add_column("S3", style="cyan")

    for name, endpoint in config.endpoints.items():
        display_name = Text(f"* {name}", style="bold yellow") if name == config.selected_endpoint else Text(f"  {name}")
        table.add_row(display_name, endpoint.api, endpoint.s3)
    return table


@dataclass
class SharedState:
    log_file: Optional[Path] = None
    verbose: bool = False
    debug: bool = False
    max_table_size: int = MAX_TABLE_SIZE


SHARED_STATE = SharedState()


def get_shared_state() -> SharedState:
    return SHARED_STATE



from __future__ import annotations

import base64
import hashlib
import math
import re
import string
import traceback
from hashlib import md5
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from uuid import UUID

import yaml
from rich.console import Console

from kleinkram._version import __version__
from kleinkram.errors import FileNameNotSupported
from kleinkram.errors import FileTypeNotSupported
from kleinkram.models import File
from kleinkram.types import IdLike
from kleinkram.types import PathLike

INTERNAL_ALLOWED_CHARS = string.ascii_letters + string.digits + "_" + "-"
SUPPORT_FILE_TYPES = [".bag", ".mcap", ".db3", ".svo2", ".tum", ".yaml", ".yml"]
EXPERIMENTAL_FILE_TYPES = []


def sanitize_path_component(name: str) -> str:
    """Sanitize a path component to prevent path traversal."""
    name = name.replace("/", "_").replace("\\", "_")
    if name in (".", ".."):
        name = name.replace(".", "_")
    return name


def file_paths_from_files(files: Sequence[File], *, dest: Path, allow_nested: bool = False) -> Dict[Path, File]:
    """\
    determines the destinations for a sequence of `File` objects,
    possibly nested by project and mission
    """
    if len(set([file.mission_id for file in files])) > 1 and not allow_nested:
        raise ValueError("files from multiple missions were selected")

    result = {}
    base_dir = dest.resolve()

    for file in files:
        file_name = sanitize_path_component(file.name)
        if not allow_nested:
            path = dest / file_name
        else:
            project_name = sanitize_path_component(file.project_name)
            mission_name = sanitize_path_component(file.mission_name)
            path = dest / project_name / mission_name / file_name

        resolved_path = path.resolve()
        try:
            resolved_path.relative_to(base_dir)
        except ValueError:
            raise ValueError(f"Refusing to download to path outside the destination directory: {resolved_path}")

        result[path] = file

    return result


def upper_camel_case_to_words(s: str) -> List[str]:
    """split `s` given upper camel case to words"""
    return re.sub("([a-z])([A-Z])", r"\1 \2", s).split()


def split_args(args: Sequence[str]) -> Tuple[List[UUID], List[str]]:
    """\
    split a sequece of strings into a list of UUIDs and a list of names
    depending on whether the string is a valid UUID or not
    """
    uuids = []
    names = []
    for arg in args:
        if is_valid_uuid4(arg):
            uuids.append(UUID(arg, version=4))
        else:
            names.append(arg)
    return uuids, names


def check_file_paths(files: Sequence[Path]) -> None:
    """\
    checks that files exist, are files and have a supported file suffix

    NOTE: kleinkram treats filesuffixes as filetypes and limits
    the supported suffixes
    """
    for file in files:
        check_file_path(file)


def check_file_path(file: Path) -> None:
    if file.is_dir():
        raise FileNotFoundError(f"{file} is a directory and not a file")
    if not file.exists():
        raise FileNotFoundError(f"{file} does not exist")
    if file.suffix not in SUPPORT_FILE_TYPES:
        raise FileTypeNotSupported(f"only {', '.join(SUPPORT_FILE_TYPES)} files are supported: {file}")
    if not check_filename_is_sanatized(file.stem):
        raise FileNameNotSupported(
            f"only `{''.join(INTERNAL_ALLOWED_CHARS)}` are " f"allowed in filenames and at most 50chars: {file}"
        )


def format_error(msg: str, exc: Exception, *, verbose: bool = False) -> str:
    if not verbose:
        ret = f"{msg}: {type(exc).__name__}"
    else:
        ret = f"{msg}: {exc}"
    return styled_string(ret, style="red")


def format_traceback(exc: Exception) -> str:
    return "".join(traceback.format_exception(type(exc), value=exc, tb=exc.__traceback__))


def styled_string(*objects: Any, **kwargs: Any) -> str:
    """\
    accepts any object that Console.print can print
    returns the raw string output
    """
    console = Console()
    with console.capture() as capture:
        console.print(*objects, **kwargs, end="")
    return capture.get()


def is_valid_uuid4(uuid: str) -> bool:
    try:
        UUID(uuid, version=4)
        return True
    except ValueError:
        return False


def check_filename_is_sanatized(filename: str) -> bool:
    if len(filename) > 50:
        return False
    if not all(char in INTERNAL_ALLOWED_CHARS for char in filename):
        return False
    return True


def get_filename(path: Path) -> str:
    """\
    takes a path and returns a sanitized filename

    the format for this internal filename is:
    - replace all disallowed characters with "_"
    - trim to 40 chars + 10 hashed chars
        - the 10 hashed chars are deterministic given the original filename
    """

    stem = "".join(char if char in INTERNAL_ALLOWED_CHARS else "_" for char in path.stem)

    if len(stem) > 50:
        hash = md5(path.name.encode()).hexdigest()
        stem = f"{stem[:40]}{hash[:10]}"

    return f"{stem}{path.suffix}"


def get_filename_map(file_paths: Sequence[Path]) -> Dict[str, Path]:
    """\
    takes a list of unique filepaths and returns a mapping
    from the original filename to a sanitized internal filename
    see `get_filename` for the internal filename format
    """

    if len(file_paths) != len(set(file_paths)):
        raise ValueError("files paths must be unique")

    internal_file_map = {}
    for file in file_paths:
        if file.is_dir():
            raise ValueError(f"got dir {file} expected file")

        internal_file_map[get_filename(file)] = file

    if len(internal_file_map) != len(set(internal_file_map.values())):
        raise RuntimeError("hash collision")

    return internal_file_map


def b64_md5(file: Path) -> str:
    hash_md5 = hashlib.md5()
    with open(file, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    binary_digest = hash_md5.digest()
    return base64.b64encode(binary_digest).decode("utf-8")


def load_metadata(path: Path) -> Dict[str, str]:
    if not path.exists():
        raise FileNotFoundError(f"metadata file not found: {path}")
    try:
        with path.open() as f:
            return {str(k): str(v) for k, v in yaml.safe_load(f).items()}
    except Exception as e:
        raise ValueError(f"could not parse metadata file: {e}")


def get_supported_api_version() -> Tuple[int, int, int]:
    vers = __version__.split(".")
    return tuple(map(int, vers[:3]))  # type: ignore


T = TypeVar("T")


def singleton_list(x: Optional[T]) -> List[T]:
    return [] if x is None else [x]


def parse_uuid_like(s: IdLike) -> UUID:
    return UUID(str(s))


def parse_path_like(s: PathLike) -> Path:
    return Path(s)


def format_bytes(size_bytes: int | float, speed: bool = False) -> str:
    """Formats a size in bytes into a human-readable string with appropriate units."""
    if size_bytes == 0:
        return "0 B/s" if speed else "0 B"

    units = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    power = math.floor(math.log(size_bytes, 1000))
    unit_index = min(power, len(units) - 1)

    value = size_bytes / (1000**unit_index)

    unit_suffix = "/s" if speed else ""
    return f"{value:.2f} {units[unit_index]}{unit_suffix}"



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Mapping
from typing import Optional
from typing import cast

from kleinkram.api.client import AuthenticatedClient

DataPage = Dict[str, Any]


PAGE_SIZE = 128
SKIP = "skip"
TAKE = "take"
EXACT_MATCH = "exactMatch"


def paginated_request(
    client: AuthenticatedClient,
    endpoint: str,
    params: Optional[Mapping[str, Any]] = None,
    max_entries: Optional[int] = None,
    page_size: int = PAGE_SIZE,
    exact_match: bool = False,
) -> Generator[DataPage, None, None]:
    total_entries_count = 0

    params = dict(params or {})

    params[TAKE] = page_size
    params[SKIP] = 0
    if exact_match:
        params[EXACT_MATCH] = str(exact_match).lower()  # pass string rather than bool

    while True:
        resp = client.get(endpoint, params=params)

        # explicitly handle 404 if json contains message
        if resp.status_code == 404 and "message" in resp.json():
            raise ValueError(resp.json()["message"])

        # raise for other errors
        resp.raise_for_status()

        paged_data = resp.json()
        data_page = cast(List[DataPage], paged_data["data"])

        for entry in data_page:
            total_entries_count += 1
            yield entry
            if max_entries is not None and max_entries <= total_entries_count:
                return

        count = cast(int, paged_data["count"])
        skip = cast(int, paged_data["skip"])
        take = cast(int, paged_data["take"])

        if count - skip - take <= 0:
            return

        params[SKIP] = total_entries_count



from __future__ import annotations

import subprocess
import sys
import time
from pathlib import Path
from secrets import token_hex

import pytest

from kleinkram import create_mission
from kleinkram import create_project
from kleinkram import delete_project
from kleinkram import list_missions
from kleinkram import list_projects
from kleinkram import upload

# we expect the mission files to be in this folder that is not commited to the repo
DATA_PATH = Path(__file__).parent / "data"
DATA_FILES = [
    DATA_PATH / "10_KB.bag",
    DATA_PATH / "50_KB.bag",
    DATA_PATH / "1_MB.bag",
    DATA_PATH / "17_MB.bag",
    DATA_PATH / "125_MB.bag",
]

PROJECT_DESCRIPTION = "This is a test project"


WAIT_BEFORE_DELETION = 5


@pytest.fixture(scope="session")
def project():
    project_name = token_hex(8)

    create_project(project_name, description="This is a test project")
    project = list_projects(project_names=[project_name])[0]

    yield project

    time.sleep(WAIT_BEFORE_DELETION)
    delete_project(project.id)


@pytest.fixture
def mission(project):
    mission_name = token_hex(8)
    upload(
        mission_name=mission_name,
        project_id=project.id,
        files=DATA_FILES,
        create=True,
    )
    mission = list_missions(project_ids=[project.id], mission_names=[mission_name])[0]

    yield mission


@pytest.fixture
def empty_mission(project):
    mission_name = token_hex(8)
    create_mission(mission_name, project.id, metadata={})
    mission = list_missions(project_ids=[project.id], mission_names=[mission_name])[0]

    yield mission


@pytest.fixture(scope="session", autouse=True)
def auto_login():
    """
    Automatically logs in using the CLI before running any tests.
    """
    try:
        # Run: python3 -m kleinkram login --user 1
        result = subprocess.run(
            [sys.executable, "-m", "kleinkram", "login", "--user", "1"],
            cwd=str(Path(__file__).parent.parent),  # Run from cli root
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            pytest.fail(
                f"Failed to auto-login. Return code: {result.returncode}\nStdout: {result.stdout}\nStderr: {result.stderr}"
            )

    except Exception as e:
        pytest.fail(f"Failed to auto-login: {e}")



from __future__ import annotations

import logging
from pathlib import Path
from typing import List
from typing import Optional

import typer

import kleinkram.core
from kleinkram.api.client import AuthenticatedClient
from kleinkram.cli._file_validator import FileValidator
from kleinkram.cli._file_validator import _report_skipped_files
from kleinkram.cli._upload import _build_mission_query
from kleinkram.config import get_shared_state
from kleinkram.printing import print_file_verification_status

logger = logging.getLogger(__name__)

HELP = """\
Verify if files were uploaded correctly.
"""

verify_typer = typer.Typer(name="verify", invoke_without_command=True, help=HELP)


def _handle_no_files_to_process(original_count: int, processed_count: int, action: str = "verify") -> None:
    """Checks if any files are left and exits if not."""
    if processed_count > 0:
        return

    if original_count > 0:
        typer.echo(
            typer.style(f"All paths were skipped. No files to {action}.", fg=typer.colors.RED),
            err=True,
        )
    else:
        typer.echo(
            typer.style(f"No files provided to {action}.", fg=typer.colors.RED),
            err=True,
        )
    raise typer.Exit(code=1)


@verify_typer.callback()
def verify(
    files: List[str] = typer.Argument(help="files to verify"),
    project: Optional[str] = typer.Option(None, "--project", "-p", help="project id or name"),
    mission: str = typer.Option(..., "--mission", "-m", help="mission id or name"),
    skip: bool = typer.Option(
        False,
        "--skip",
        "-s",
        help="skip unsupported file types, badly named files, or directories instead of erroring",
    ),
    experimental_datatypes: bool = typer.Option(False, help="allow experimental datatypes (yaml, svo2, db3, tum)"),
    skip_hash: bool = typer.Option(None, help="skip hash check"),
    check_file_hash: bool = typer.Option(
        True,
        help="check file hash. If True, file names and file hashes are checked.",
    ),
    check_file_size: bool = typer.Option(
        True,
        help="check file size. If True, file names and file sizes are checked.",
    ),
) -> None:
    # get all filepaths
    original_file_paths = [Path(file) for file in files]

    # get mission query
    mission_query = _build_mission_query(mission, project)

    validator = FileValidator(
        skip=skip,
        experimental_datatypes=experimental_datatypes,
    )
    files_to_verify = validator.filter_files(original_file_paths)

    # Report skipped files (if any)
    _report_skipped_files(validator.skipped_files)

    # Check if we have anything left to do
    _handle_no_files_to_process(
        original_count=len(original_file_paths),
        processed_count=len(files_to_verify),
        action="verify",
    )

    verbose = get_shared_state().verbose
    file_status = kleinkram.core.verify(
        client=AuthenticatedClient(),
        query=mission_query,
        file_paths=files_to_verify,
        skip_hash=skip_hash,
        check_file_hash=check_file_hash,
        check_file_size=check_file_size,
        verbose=verbose,
    )
    print_file_verification_status(file_status, pprint=verbose)



from __future__ import annotations

LOGIN_MESSAGE = "Please login using `klein login`."
UPDATE_MESSAGE = "Please update your CLI using `pip install --upgrade kleinkram`."


class ParsingError(Exception): ...


class InvalidFileQuery(Exception): ...


class InvalidMissionQuery(Exception): ...


class InvalidProjectQuery(Exception): ...


class MissionExists(Exception): ...


class ProjectExists(Exception): ...


class MissionNotFound(Exception): ...


class ProjectNotFound(Exception): ...


class FileNotFound(Exception): ...


class AccessDenied(Exception): ...


class InvalidCLIVersion(Exception): ...


class FileTypeNotSupported(Exception): ...


class FileNameNotSupported(Exception): ...


class DatatypeNotSupported(Exception): ...


class InvalidMissionMetadata(Exception): ...


class MissionValidationError(Exception): ...


class ProjectValidationError(Exception): ...


class NotAuthenticated(Exception):
    def __init__(self) -> None:
        super().__init__(LOGIN_MESSAGE)


class UpdateCLIVersion(Exception):
    def __init__(self) -> None:
        super().__init__(UPDATE_MESSAGE)


class RunNotFound(Exception): ...



"""\
this file contains dataclasses for specifying remote resources on kleinkram
here we also provide some helper functions to validate certain properties
of these specifications

additionally we provide wrappers around the api for fetching the specified
resources (TODO: move this part to another file)
"""

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import List
from uuid import UUID

from kleinkram.errors import InvalidMissionQuery
from kleinkram.errors import InvalidProjectQuery

MAX_PARALLEL_REQUESTS = 32
SPECIAL_PATTERN_CHARS = ["*", "?", "[", "]"]


@dataclass
class ProjectQuery:
    patterns: List[str] = field(default_factory=list)
    ids: List[UUID] = field(default_factory=list)


@dataclass
class MissionQuery:
    patterns: List[str] = field(default_factory=list)
    ids: List[UUID] = field(default_factory=list)
    project_query: ProjectQuery = field(default_factory=ProjectQuery)


@dataclass
class FileQuery:
    patterns: List[str] = field(default_factory=list)
    ids: List[UUID] = field(default_factory=list)
    mission_query: MissionQuery = field(default_factory=MissionQuery)


@dataclass
class RunQuery:
    mission_ids: List[UUID] = field(default_factory=list)
    mission_patterns: List[str] = field(default_factory=list)
    project_ids: List[UUID] = field(default_factory=list)
    project_patterns: List[str] = field(default_factory=list)


def check_mission_query_is_creatable(query: MissionQuery) -> str:
    """\
    check if a query is unique and can be used to create a mission
    returns: the mission name
    """
    if not mission_query_is_unique(query):
        raise InvalidMissionQuery(f"Mission query is not unique: {query}")
    # cant create a missing by id
    if query.ids:
        raise InvalidMissionQuery(f"cant create mission by id: {query}")
    return query.patterns[0]


def check_project_query_is_creatable(query: ProjectQuery) -> str:
    if not project_query_is_unique(query):
        raise InvalidProjectQuery(f"Project query is not unique: {query}")
    # cant create a missing by id
    if query.ids:
        raise InvalidProjectQuery(f"cant create project by id: {query}")
    return query.patterns[0]


def _pattern_is_unique(pattern: str) -> bool:
    for char in SPECIAL_PATTERN_CHARS:
        if char in pattern:
            return False
    return True


def project_query_is_unique(query: ProjectQuery) -> bool:
    # a single project id is specified
    if len(query.ids) == 1 and not query.patterns:
        return True

    # a single project name is specified
    if len(query.patterns) == 1 and _pattern_is_unique(query.patterns[0]):
        return True
    return False


def mission_query_is_unique(query: MissionQuery) -> bool:
    # a single mission id is specified
    if len(query.ids) == 1 and not query.patterns:
        return True

    # a single mission name a unique project spec are specified
    if project_query_is_unique(query.project_query) and len(query.patterns) == 1 and _pattern_is_unique(query.patterns[0]):
        return True
    return False


def file_query_is_unique(query: FileQuery) -> bool:
    # a single file id is specified
    if len(query.ids) == 1 and not query.patterns:
        return True

    # a single file name a unique mission spec are specified
    if mission_query_is_unique(query.mission_query) and len(query.patterns) == 1 and _pattern_is_unique(query.patterns[0]):
        return True
    return False



from __future__ import annotations

import logging
from pathlib import Path
from typing import List
from typing import Optional

import typer

import kleinkram.core
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import FileQuery
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.config import get_shared_state
from kleinkram.utils import split_args

logger = logging.getLogger(__name__)

HELP = """\
Download files from kleinkram.
"""


download_typer = typer.Typer(name="download", no_args_is_help=True, invoke_without_command=True, help=HELP)


@download_typer.callback()
def download(
    files: Optional[List[str]] = typer.Argument(None, help="file names, ids or patterns"),
    projects: Optional[List[str]] = typer.Option(None, "--project", "-p", help="project names, ids or patterns"),
    missions: Optional[List[str]] = typer.Option(None, "--mission", "-m", help="mission names, ids or patterns"),
    dest: str = typer.Option(prompt="destination", help="local path to save the files"),
    nested: bool = typer.Option(False, help="save files in nested directories, project-name/mission-name"),
    overwrite: bool = typer.Option(
        False,
        help="overwrite files if they already exist and don't match the file size or file hash",
    ),
) -> None:
    # create destination directory
    dest_dir = Path(dest)
    if not dest_dir.exists():
        typer.confirm(f"Destination {dest_dir} does not exist. Create it?", abort=True)
    dest_dir.mkdir(parents=True, exist_ok=True)

    # get file query
    file_ids, file_patterns = split_args(files or [])
    mission_ids, mission_patterns = split_args(missions or [])
    project_ids, project_patterns = split_args(projects or [])

    project_query = ProjectQuery(patterns=project_patterns, ids=project_ids)
    mission_query = MissionQuery(
        patterns=mission_patterns,
        ids=mission_ids,
        project_query=project_query,
    )
    file_query = FileQuery(patterns=file_patterns, ids=file_ids, mission_query=mission_query)

    kleinkram.core.download(
        client=AuthenticatedClient(),
        query=file_query,
        base_dir=dest_dir,
        nested=nested,
        overwrite=overwrite,
        verbose=get_shared_state().verbose,
    )



from __future__ import annotations

from typing import List
from typing import Optional

import typer

from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import FileQuery
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.api.routes import get_files
from kleinkram.api.routes import get_missions
from kleinkram.api.routes import get_projects
from kleinkram.config import get_shared_state
from kleinkram.printing import print_files
from kleinkram.printing import print_missions
from kleinkram.printing import print_projects
from kleinkram.utils import split_args

HELP = """\
List projects, missions, or files.
"""


list_typer = typer.Typer(name="list", invoke_without_command=True, help=HELP, no_args_is_help=True)


@list_typer.command()
def files(
    files: Optional[List[str]] = typer.Argument(
        None,
        help="file names, ids or patterns",
    ),
    projects: Optional[List[str]] = typer.Option(None, "--project", "-p", help="project name or id"),
    missions: Optional[List[str]] = typer.Option(None, "--mission", "-m", help="mission name or id"),
) -> None:
    file_ids, file_patterns = split_args(files or [])
    mission_ids, mission_patterns = split_args(missions or [])
    project_ids, project_patterns = split_args(projects or [])

    project_query = ProjectQuery(patterns=project_patterns, ids=project_ids)
    mission_query = MissionQuery(
        project_query=project_query,
        ids=mission_ids,
        patterns=mission_patterns,
    )
    file_query = FileQuery(mission_query=mission_query, patterns=file_patterns, ids=file_ids)

    client = AuthenticatedClient()
    parsed_files = list(get_files(client, file_query=file_query))
    print_files(parsed_files, pprint=get_shared_state().verbose)


@list_typer.command()
def missions(
    projects: Optional[List[str]] = typer.Option(None, "--project", "-p", help="project name or id"),
    missions: Optional[List[str]] = typer.Argument(None, help="mission names"),
) -> None:
    mission_ids, mission_patterns = split_args(missions or [])
    project_ids, project_patterns = split_args(projects or [])

    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)
    mission_query = MissionQuery(
        ids=mission_ids,
        patterns=mission_patterns,
        project_query=project_query,
    )

    client = AuthenticatedClient()
    parsed_missions = list(get_missions(client, mission_query=mission_query))
    print_missions(parsed_missions, pprint=get_shared_state().verbose)


@list_typer.command()
def projects(
    projects: Optional[List[str]] = typer.Argument(None, help="project names"),
) -> None:
    project_ids, project_patterns = split_args(projects or [])
    project_query = ProjectQuery(patterns=project_patterns, ids=project_ids)

    client = AuthenticatedClient()
    parsed_projects = list(get_projects(client, project_query=project_query))
    print_projects(parsed_projects, pprint=get_shared_state().verbose)



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import List

import typer

from kleinkram.config import get_shared_state
from kleinkram.errors import DatatypeNotSupported
from kleinkram.errors import FileNameNotSupported
from kleinkram.utils import EXPERIMENTAL_FILE_TYPES
from kleinkram.utils import SUPPORT_FILE_TYPES
from kleinkram.utils import check_filename_is_sanatized


@dataclass
class FileValidator:
    """Encapsulates all file validation logic based on CLI flags."""

    skip: bool
    experimental_datatypes: bool

    # Stores (file, reason) for skipped files
    skipped_files: List[tuple[Path, str]] = field(default_factory=list)

    def filter_files(self, file_paths: List[Path]) -> List[Path]:
        """
        Validates a list of file paths.

        - Populates `self.skipped_files` with invalid files.
        - Raises an exception on the first invalid file if `self.skip` is False.
        - Returns a list of valid files to verify.
        """
        files_to_verify = []
        for file in file_paths:
            try:
                self._validate_path(file)
                files_to_verify.append(file)
            except (
                FileNotFoundError,
                IsADirectoryError,
                DatatypeNotSupported,
                FileNameNotSupported,
            ) as e:
                if self.skip:
                    self.skipped_files.append((file, str(e)))
                else:
                    # Re-raise the exception to be caught by Typer
                    self._raise_with_skip_hint(e)
        return files_to_verify

    def _validate_path(self, file: Path) -> None:
        """
        Runs a single file through all validation checks.
        Raises an error if any check fails.
        """
        # 0. Check for existence
        if not file.exists():
            raise FileNotFoundError(f"File not found: {file}")

        # 1. Check for directories
        if file.is_dir():
            raise IsADirectoryError(f"{file} is a directory")

        file_suffix = file.suffix.lower()
        is_experimental = file_suffix in EXPERIMENTAL_FILE_TYPES
        is_standard = file_suffix in SUPPORT_FILE_TYPES

        # 2. Check if the datatype is known at all
        if not is_standard and not is_experimental:
            raise DatatypeNotSupported(f"Unsupported file type '{file_suffix}' on file {file}")

        # 3. Check if an experimental datatype is allowed
        if is_experimental and not self.experimental_datatypes:
            raise DatatypeNotSupported(f"Experimental datatype '{file_suffix}' not enabled for file {file}")

        # 4. Check filename
        is_bad_name = not check_filename_is_sanatized(file.stem)
        if is_bad_name:
            raise FileNameNotSupported(f"Badly formed filename for file {file}")

    def _raise_with_skip_hint(self, e: Exception) -> None:
        """Re-raises a validation error with a hint to use --skip."""
        base_message = str(e)
        hint = "Use --skip to ignore."

        if isinstance(e, IsADirectoryError):
            hint = "Use --skip to ignore directories."
        elif isinstance(e, DatatypeNotSupported):
            if "Experimental" in base_message:
                hint = "Use --experimental-datatypes to allow or --skip to ignore."
        elif isinstance(e, FileNameNotSupported):
            hint = "Use --skip to ignore."  # No --fix-filenames hint here

        # Raise a new exception to the same type to preserve the error type
        raise type(e)(f"{base_message}. {hint}")


def _report_skipped_files(skipped_files: List[tuple[Path, str]]) -> None:
    """Prints a formatted report of all skipped files."""
    if skipped_files:
        debug = get_shared_state().debug

        if debug:
            # Full report
            typer.echo(
                typer.style(
                    f"--- Skipped {len(skipped_files)} path(s) ---",
                    fg=typer.colors.YELLOW,
                )
            )
            for file, reason in skipped_files:
                typer.echo(f"Skipped: {file} (Reason: {reason})")
            typer.echo("---------------------------\n")
        else:
            # Summary report
            typer.echo(
                typer.style(
                    f"Skipped {len(skipped_files)} path(s) due to errors. Use `klein --debug [...]` for details.",
                    fg=typer.colors.YELLOW,
                )
            )
            typer.echo("")



from __future__ import annotations

import json
import sys
import time
from dataclasses import asdict
from datetime import datetime
from datetime import timezone
from pathlib import Path
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import dateutil.parser
import httpx
import typer
from rich.console import Console
from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.spinner import Spinner
from rich.table import Table
from rich.text import Text

import kleinkram
from kleinkram.api.client import AuthenticatedClient
from kleinkram.config import get_shared_state
from kleinkram.core import FileVerificationStatus
from kleinkram.models import ActionTemplate
from kleinkram.models import File
from kleinkram.models import FileState
from kleinkram.models import LogEntry
from kleinkram.models import MetadataValue
from kleinkram.models import MetadataValueType
from kleinkram.models import Mission
from kleinkram.models import Project
from kleinkram.models import Run

FILE_STATE_COLOR = {
    FileState.OK: "green",
    FileState.CORRUPTED: "red",
    FileState.UPLOADING: "yellow",
    FileState.ERROR: "red",
    FileState.CONVERTING: "blue",
    FileState.CONVERSION_ERROR: "red",
    FileState.LOST: "bold red",
    FileState.FOUND: "yellow",
}


FILE_VERIFICATION_STATUS_STYLES = {
    FileVerificationStatus.UPLOADED: "green",
    FileVerificationStatus.UPLOADING: "yellow",
    FileVerificationStatus.MISSING: "yellow",
    FileVerificationStatus.MISMATCHED_HASH: "red",
    FileVerificationStatus.MISMATCHED_SIZE: "red",
    FileVerificationStatus.UNKNOWN: "gray",
    FileVerificationStatus.COMPUTING_HASH: "purple",
}


def _add_placeholder_row(table: Table, skipped: int) -> None:
    first_column = f"... ({skipped} more)"
    table.add_row(first_column, *["..." for _ in range(len(table.columns) - 1)])


def file_state_to_text(file_state: FileState) -> Text:
    return Text(file_state.value, style=FILE_STATE_COLOR[file_state])


def file_verification_status_to_text(
    file_verification_status: FileVerificationStatus,
) -> Text:
    return Text(
        file_verification_status.value,
        style=FILE_VERIFICATION_STATUS_STYLES[file_verification_status],
    )


def format_bytes(size: int) -> str:
    """
    Converts a size in bytes to a human-readable string (e.g., KB, MB, GB, TB).

    Args:
        size (int): The size in bytes.

    Returns:
        str: A formatted string representing the size in appropriate units.
    """

    if size < 0:
        raise ValueError("Size must be a non-negative integer")

    # Define units and their thresholds
    units = ["B", "KB", "MB", "GB", "TB", "PB", "EB"]
    index = 0

    fsize: float = size
    while fsize >= 1000 and index < len(units) - 1:
        fsize /= 1000.0
        index += 1

    # Format to 2 decimal places if needed
    if index == 0:  # Bytes don't need decimals
        return f"{int(fsize)} {units[index]}"
    else:
        return f"{fsize:.2f} {units[index]}"


MetadataValueInternalType = Union[str, float, bool, datetime]


def parse_metadata_value(value: MetadataValue) -> Union[str, float, bool, datetime]:
    if value.type_ in [
        MetadataValueType.STRING,
        MetadataValueType.LINK,
        MetadataValueType.LOCATION,
    ]:
        return value.value
    if value.type_ == MetadataValueType.NUMBER:
        return float(value.value)
    if value.type_ == MetadataValueType.BOOLEAN:
        return value.value == "true"
    if value.type_ == MetadataValueType.DATE:
        return dateutil.parser.isoparse(value.value)
    assert False, "unreachable"


def projects_to_table(projects: Sequence[Project]) -> Table:
    table = Table(title="projects")
    table.add_column("id")
    table.add_column("name")
    table.add_column("description")

    max_table_size = get_shared_state().max_table_size
    for project in projects[:max_table_size]:
        table.add_row(str(project.id), project.name, project.description)
    if len(projects) > max_table_size:
        _add_placeholder_row(table, skipped=len(projects) - max_table_size)
    return table


def missions_to_table(missions: Sequence[Mission]) -> Table:
    table = Table(title="missions")
    table.add_column("project")
    table.add_column("name")
    table.add_column("id")
    table.add_column("files")
    table.add_column("size")

    # order by project, name
    missions_tp: List[Tuple[str, str, Mission]] = []
    for mission in missions:
        missions_tp.append((mission.project_name, mission.name, mission))
    missions_tp.sort()

    if not missions_tp:
        return table
    last_project: Optional[str] = None
    max_table_size = get_shared_state().max_table_size
    for project, _, mission in missions_tp[:max_table_size]:
        # add delimiter row if project changes
        if last_project is not None and last_project != project:
            table.add_section()
        last_project = project

        table.add_row(
            mission.project_name,
            mission.name,
            str(mission.id),
            str(mission.number_of_files),
            format_bytes(mission.size),
        )

    if len(missions_tp) > max_table_size:
        _add_placeholder_row(table, skipped=len(missions_tp) - max_table_size)
    return table


def files_to_table(files: Sequence[File], *, title: str = "files", delimiters: bool = True) -> Table:
    table = Table(title=title)
    table.add_column("project")
    table.add_column("mission")
    table.add_column("name")
    table.add_column("id")
    table.add_column("state")
    table.add_column("size")
    table.add_column("categories")

    # order by project, mission, name
    files_tp: List[Tuple[str, str, str, File]] = []
    for file in files:
        files_tp.append((file.project_name, file.mission_name, file.name, file))
    files_tp.sort()

    if not files_tp:
        return table

    last_mission: Optional[str] = None
    max_table_size = get_shared_state().max_table_size
    for _, mission, _, file in files_tp[:max_table_size]:
        if last_mission is not None and last_mission != mission and delimiters:
            table.add_section()
        last_mission = mission

        table.add_row(
            file.project_name,
            file.mission_name,
            file.name,
            Text(str(file.id), style="green"),
            file_state_to_text(file.state),
            format_bytes(file.size),
            ", ".join(file.categories),
        )

    if len(files_tp) > max_table_size:
        _add_placeholder_row(table, skipped=len(files_tp) - max_table_size)

    return table


def file_info_table(file: File) -> Table:
    table = Table("k", "v", title=f"file info: {file.name}", show_header=False)

    table.add_row("name", file.name)
    table.add_row("id", Text(str(file.id), style="green"))
    table.add_row("project", file.project_name)
    table.add_row("project id", Text(str(file.project_id), style="green"))
    table.add_row("mission", file.mission_name)
    table.add_row("mission id", Text(str(file.mission_id), style="green"))
    table.add_row("created", str(file.created_at))
    table.add_row("updated", str(file.updated_at))
    table.add_row("size", format_bytes(file.size))
    table.add_row("state", file_state_to_text(file.state))
    table.add_row("categories", ", ".join(file.categories))
    table.add_row("topics", ", ".join(file.topics))
    table.add_row("hash", file.hash)
    table.add_row("type", file.type_)
    table.add_row("date", str(file.date))

    return table


def mission_info_table(mission: Mission, print_metadata: bool = True) -> Tuple[Table, ...]:
    table = Table("k", "v", title=f"mission info: {mission.name}", show_header=False)

    # TODO: add more fields as we store more information in the Mission object
    table.add_row("name", mission.name)
    table.add_row("id", Text(str(mission.id), style="green"))
    table.add_row("project", mission.project_name)
    table.add_row("project id", Text(str(mission.project_id), style="green"))
    table.add_row("created", str(mission.created_at))
    table.add_row("updated", str(mission.updated_at))
    table.add_row("size", format_bytes(mission.size))
    table.add_row("files", str(mission.number_of_files))

    if not print_metadata:
        return (table,)

    metadata_table = Table("k", "v", title="mission metadata", show_header=False)
    kv_pairs_sorted = sorted([(k, v) for k, v in mission.metadata.items()], key=lambda x: x[0])
    for k, v in kv_pairs_sorted:
        metadata_table.add_row(k, str(parse_metadata_value(v)))

    return table, metadata_table


def project_info_table(project: Project) -> Table:
    table = Table("k", "v", title=f"project info: {project.name}", show_header=False)

    # TODO: add more fields as we store more information in the Project object
    table.add_row("id", Text(str(project.id), style="green"))
    table.add_row("name", project.name)
    table.add_row("description", project.description)
    table.add_row("created", str(project.created_at))
    table.add_row("updated", str(project.updated_at))
    table.add_row("required tags", ", ".join(project.required_tags))

    return table


def file_verification_status_table(
    file_status: Mapping[Path, FileVerificationStatus],
) -> Table:
    table = Table(title="file status")
    table.add_column("filename", style="cyan")
    table.add_column("status", style="green")
    for path, status in file_status.items():
        table.add_row(str(path), file_verification_status_to_text(status))
    return table


def print_file_verification_status(file_status: Mapping[Path, FileVerificationStatus], *, pprint: bool) -> None:
    """\
    prints the file verification status to stdout / stderr
    either using pprint or as a list for piping
    """
    if pprint:
        table = file_verification_status_table(file_status)
        Console().print(table)
    else:
        for path, status in file_status.items():
            stream = sys.stdout if status == FileVerificationStatus.UPLOADED else sys.stderr
            print(path, file=stream, flush=True)


def print_files(files: Sequence[File], *, pprint: bool) -> None:
    """\
    prints the files to stdout / stderr
    either using pprint or as a list for piping
    """
    if pprint:
        table = files_to_table(files)
        Console().print(table)
    else:
        for file in files:
            stream = sys.stdout if file.state == FileState.OK else sys.stderr
            print(file.id, file=stream, flush=True)


def print_missions(missions: Sequence[Mission], *, pprint: bool) -> None:
    """\
    prints the missions to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        table = missions_to_table(missions)
        Console().print(table)
    else:
        for mission in missions:
            print(mission.id)


def print_projects(projects: Sequence[Project], *, pprint: bool) -> None:
    """\
    prints the projects to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        table = projects_to_table(projects)
        Console().print(table)
    else:
        for project in projects:
            print(project.id)


def print_file_info(file: File, *, pprint: bool) -> None:
    """\
    prints the file info to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        table = file_info_table(file)
        Console().print(table)
    else:
        file_dct = asdict(file)
        for key in file_dct:
            file_dct[key] = str(file_dct[key])  # TODO: improve this
        print(json.dumps(file_dct))


def print_mission_info(mission: Mission, *, pprint: bool) -> None:
    """\
    prints the mission info to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        Console().print(*mission_info_table(mission, print_metadata=True))
    else:
        mission_dct = asdict(mission)
        for key in mission_dct:
            mission_dct[key] = str(mission_dct[key])  # TODO: improve this
        print(json.dumps(mission_dct))


def print_project_info(project: Project, *, pprint: bool) -> None:
    """\
    prints the project info to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        Console().print(project_info_table(project))
    else:
        project_dct = asdict(project)
        for key in project_dct:
            project_dct[key] = str(project_dct[key])  # TODO: improve this
        print(json.dumps(project_dct))


def runs_to_table(runs: Sequence[Run]) -> Table:
    table = Table(title="action runs")
    table.add_column("project")
    table.add_column("mission")
    table.add_column("template")
    table.add_column("run id")
    table.add_column("status")
    table.add_column("created")

    # order by created_at descending
    runs_sorted = sorted(runs, key=lambda r: r.created_at, reverse=True)

    max_table_size = get_shared_state().max_table_size
    for run in runs_sorted[:max_table_size]:
        table.add_row(
            run.project_name,
            run.mission_name,
            run.template_name,
            Text(str(run.uuid), style="green"),
            run.state,
            str(run.created_at),
        )

    if len(list(runs)) > max_table_size:
        _add_placeholder_row(table, skipped=len(runs) - max_table_size)
    return table


def run_info_table(run: Run) -> Table:
    table = Table("k", "v", title=f"run info: {run.uuid}", show_header=False)

    table.add_row("id", Text(str(run.uuid), style="green"))
    table.add_row("template", run.template_name)
    table.add_row("status", run.state)
    table.add_row("project", run.project_name)
    table.add_row("mission", run.mission_name)
    table.add_row("created", str(run.created_at))

    finished = str(run.updated_at) if run.updated_at else "N/A"
    table.add_row("updated", finished)

    return table


def print_runs_table(runs: Sequence[Run], *, pprint: bool) -> None:
    """
    Prints the runs to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        table = runs_to_table(runs)
        Console().print(table)
    else:
        for run in runs:
            print(run.uuid)


def print_run_info(run: Run, *, pprint: bool) -> None:
    """
    Prints the run info to stdout
    either using pprint or as JSON for piping
    """
    if pprint:
        Console().print(run_info_table(run))
    else:
        run_dict = asdict(run)
        for key in run_dict:
            run_dict[key] = str(run_dict[key])  # simple serialization
        print(json.dumps(run_dict))


LOG_LEVEL_COLORS = {
    "DEBUG": typer.colors.CYAN,
    "INFO": typer.colors.GREEN,
    "WARNING": typer.colors.YELLOW,
    "ERROR": typer.colors.RED,
    "CRITICAL": typer.colors.BRIGHT_RED,
    "STDOUT": typer.colors.BRIGHT_BLACK,
    "STDERR": typer.colors.RED,
}


def pretty_print_log(entry: LogEntry) -> None:
    """
    Prints a single LogEntry object to the console with
    colors and standardized formatting.

    This version correctly handles carriage returns (from tqdm)
    and empty lines.
    """
    # Clean up the level name, just in case
    level = entry.level.upper().strip()
    color = LOG_LEVEL_COLORS.get(level, typer.colors.WHITE)
    timestamp_str = entry.timestamp.strftime("%Y-%m-%d %H:%M:%S")
    level_str = f"[{level.ljust(8)}]"
    message = entry.message.strip()

    if not message:
        return

    typer.secho(f"[{timestamp_str}] {level_str} ", fg=color, nl=False)
    typer.echo(message)


def print_run_logs(logs: Sequence[LogEntry], *, pprint: bool) -> None:
    """
    Prints a sequence of LogEntry objects to the console.
    (This function is unchanged, as the logic is fully
    contained in pretty_print_log.)
    """
    if not logs:
        typer.secho("No logs found for this run.", fg=typer.colors.YELLOW)
        return

    for log_entry in logs:
        if pprint:
            pretty_print_log(log_entry)
        else:
            typer.echo(f"[{log_entry.timestamp}] {log_entry.message}")


def action_templates_to_table(templates: Sequence[ActionTemplate]) -> Table:
    """Creates a rich Table for a list of ActionTemplates."""
    table = Table(title="Available Action Templates")

    table.add_column("Name", style="cyan", no_wrap=True)
    table.add_column("ID (UUID)", style="magenta")
    table.add_column("Image Name", style="green")
    table.add_column("Command", style="cyan")

    for template in templates:
        uuid_text = Text(str(template.uuid), style="magenta")
        table.add_row(template.name, uuid_text, template.image_name, template.command)

    return table


def print_action_templates_table(templates: Sequence[ActionTemplate], *, pprint: bool) -> None:
    """
    Prints the action templates to stdout
    either using rich or as a simple list of IDs for piping.
    """
    if not templates:
        typer.echo("No action templates found.")
        return

    if pprint:
        table = action_templates_to_table(templates)
        Console().print(table)
    else:
        for template in templates:
            print(template.uuid)


def generate_live_layout(run_details: Run) -> Group:
    verbose = get_shared_state().verbose

    # Calculate elapsed time
    if run_details.updated_at:
        elapsed = run_details.updated_at - run_details.created_at
    else:
        elapsed = datetime.now(timezone.utc) - run_details.created_at

    elapsed_seconds = int(elapsed.total_seconds())
    hours, remainder = divmod(elapsed_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    elapsed_str = f"{hours}:{minutes:02}:{seconds:02}"

    state_upper = run_details.state.upper()

    if state_upper == "DONE":
        status_color = "green"
        icon = Text("[✔]", style="green")
        state_text = "DONE"
    elif state_upper in {"FAILED", "UNPROCESSABLE"}:
        status_color = "red"
        icon = Text("[✘]", style="red")
        state_text = f"{run_details.state}"
    else:
        status_color = "yellow"
        icon = Spinner("dots", style="blue")
        state_text = f"{run_details.state}"

    header_text = Text()
    header_text.append(f"Running Action {run_details.uuid} ", style="bold")
    header_text.append(f"{elapsed_str} ", style="dim")
    header_text.append(f"({state_text})", style=status_color)

    header_table = Table.grid(padding=(0, 1))
    header_table.add_column()
    header_table.add_column()
    header_table.add_row(icon, header_text)

    logs_text = Text()
    last_logs = run_details.logs[-20:]
    if not last_logs:
        logs_text.append(" => ", style="blue")
        if state_upper in {"DONE", "FAILED", "UNPROCESSABLE"}:
            logs_text.append("No logs produced.", style="dim")
        else:
            logs_text.append("Waiting for logs...", style="dim")
    else:
        for i, log in enumerate(last_logs):
            logs_text.append(" => ", style="blue")

            if verbose:
                level = log.level.upper().strip()
                color = LOG_LEVEL_COLORS.get(level, "white")
                timestamp_str = log.timestamp.strftime("%Y-%m-%d %H:%M:%S")
                level_str = f"[{level.ljust(8)}]"
                logs_text.append(f"[{timestamp_str}] {level_str} ", style=str(color))

            logs_text.append(log.message.strip())
            if i < len(last_logs) - 1:
                logs_text.append("\n")

    return Group(header_table, logs_text)


def follow_run_logs(client: AuthenticatedClient, run_uuid: str) -> int:
    """
    Polls the API for run details and prints new logs as they arrive.

    Returns:
        An exit code (0 for success, 1 for failure).
    """
    typer.echo(f"Following logs for run {run_uuid}...")

    TERMINAL_STATES = {"DONE", "FAILED", "UNPROCESSABLE"}
    current_run_state = None
    exit_code = 0  # Assume success

    try:
        with Live(refresh_per_second=10) as live:
            while current_run_state not in TERMINAL_STATES:
                try:
                    run_details: Run = kleinkram.api.routes.get_run(client, run_uuid)
                    current_run_state = run_details.state.upper()

                    live.update(generate_live_layout(run_details), refresh=True)

                    if current_run_state in TERMINAL_STATES:
                        if current_run_state != "DONE":
                            exit_code = 1  # Set failure exit code
                        break

                    time.sleep(1)  # Poll every 1 seconds for a more responsive timer

                except kleinkram.errors.RunNotFound:
                    time.sleep(1)
                except httpx.HTTPStatusError:
                    time.sleep(3)  # Wait longer on API errors
                except httpx.RequestError as e:
                    # Catch raw socket errors (like Connection reset by peer)
                    err_msg = Text(f"\nConnection lost. Please check your network and try again. ({e})", style="red")
                    live.console.print(err_msg)
                    time.sleep(3)

        # After the Live block finishes, print the final state clearly
        try:
            run_details: Run = kleinkram.api.routes.get_run(client, run_uuid)
            state_upper = run_details.state.upper()

            has_warnings = False
            for log in run_details.logs:
                if "CORRUPTED" in log.message.upper() or log.level.upper() in {"WARN", "WARNING", "ERROR"}:
                    has_warnings = True
                    break

            if state_upper == "DONE" and has_warnings:
                color = typer.colors.YELLOW
                state_display = "DONE (Completed with Warnings)"
            elif state_upper == "DONE":
                color = typer.colors.GREEN
                state_display = "DONE"
            else:
                color = typer.colors.RED
                state_display = run_details.state

            state_cause_str = f" ({run_details.state_cause})" if run_details.state_cause else ""
            typer.secho(
                f"\nRun finished with state: {state_display}{state_cause_str}",
                fg=color,
            )
        except httpx.HTTPStatusError as e:
            typer.secho(f"\nFailed to fetch final run details (API error): {e}", fg=typer.colors.RED)
        except httpx.RequestError as e:
            typer.secho(f"\nFailed to fetch final run details (network error): {e}", fg=typer.colors.RED)
        except Exception as e:
            typer.secho(f"\nFailed to fetch final run details: {e}", fg=typer.colors.RED)

    except KeyboardInterrupt:
        typer.secho(
            f"\nStopped following logs. Run {run_uuid} is still processing.",
            fg=typer.colors.YELLOW,
        )
        # Return 0, as the command itself wasn't a failure
        return 0

    return exit_code



from __future__ import annotations

import os
import struct
import time

from rosbags.rosbag1 import Writer


def serialize_string(data):
    """Serialize a string for ROS1 (std_msgs/String)."""
    encoded = data.encode("utf-8")
    return struct.pack("<I", len(encoded)) + encoded


def serialize_time(secs, nsecs):
    return struct.pack("<II", secs, nsecs)


def serialize_header(seq, secs, nsecs, frame_id):
    return struct.pack("<I", seq) + serialize_time(secs, nsecs) + serialize_string(frame_id)


def serialize_log(seq, secs, nsecs, frame_id, level, name, msg, file, function, line, topics):
    # rosgraph_msgs/Log
    # Header header
    # byte level
    # string name
    # string msg
    # string file
    # string function
    # uint32 line
    # string[] topics
    data = serialize_header(seq, secs, nsecs, frame_id)
    data += struct.pack("<b", level)
    data += serialize_string(name)
    data += serialize_string(msg)
    data += serialize_string(file)
    data += serialize_string(function)
    data += struct.pack("<I", line)
    data += struct.pack("<I", len(topics))
    for t in topics:
        data += serialize_string(t)
    return data


def serialize_temperature(seq, secs, nsecs, frame_id, temp, variance):
    # sensor_msgs/Temperature
    # Header header
    # float64 temperature
    # float64 variance
    data = serialize_header(seq, secs, nsecs, frame_id)
    data += struct.pack("<dd", temp, variance)
    return data


def serialize_time_reference(seq, secs, nsecs, frame_id, ref_secs, ref_nsecs, source):
    # sensor_msgs/TimeReference
    # Header header
    # time time_ref
    # string source
    data = serialize_header(seq, secs, nsecs, frame_id)
    data += serialize_time(ref_secs, ref_nsecs)
    data += serialize_string(source)
    return data


def serialize_twist_stamped(seq, secs, nsecs, frame_id, linear, angular):
    # geometry_msgs/TwistStamped
    # Header header
    # Twist twist (Vector3 linear, Vector3 angular)
    data = serialize_header(seq, secs, nsecs, frame_id)
    data += struct.pack("<ddd", linear[0], linear[1], linear[2])
    data += struct.pack("<ddd", angular[0], angular[1], angular[2])
    return data


def serialize_tf_message(transforms):
    # tf2_msgs/TFMessage
    # geometry_msgs/TransformStamped[] transforms
    data = struct.pack("<I", len(transforms))
    for t in transforms:
        # TransformStamped
        # Header header
        # string child_frame_id
        # Transform transform (Vector3 translation, Quaternion rotation)
        data += serialize_header(t["seq"], t["secs"], t["nsecs"], t["frame_id"])
        data += serialize_string(t["child_frame_id"])
        data += struct.pack("<ddd", t["tx"], t["ty"], t["tz"])
        data += struct.pack("<dddd", t["rx"], t["ry"], t["rz"], t["rw"])
    return data


def generate_bag(filename, target_size):
    # Adjust payload to be smaller for small files
    payload_size = 1024
    if target_size > 1024 * 1024:
        payload_size = 1024 * 1024

    if target_size < 2000:  # Very small files
        payload_size = 100

    payload = "x" * payload_size
    serialized_msg = serialize_string(payload)

    # Calculate number of messages needed
    # Approximate overhead per message in bag file is ~30-50 bytes (record header) + connection ref
    # We'll assume overhead is small compared to payload for large files, but significant for small ones.
    # We'll just write until we think we are close.

    msg_size = len(serialized_msg) + 50  # rough estimate including record headers
    num_msgs = int(target_size / msg_size) + 1
    if num_msgs < 1:
        num_msgs = 1

    print(f"Generating {filename} (~{target_size} bytes) with {num_msgs} messages of payload {payload_size}...")

    if os.path.exists(filename):
        os.remove(filename)

    with Writer(filename) as writer:
        # Add a connection
        # msg_def for std_msgs/String is just "string data"
        conn = writer.add_connection(
            topic="/test_topic",
            msgtype="std_msgs/msg/String",
            msgdef="string data",
            md5sum="992ce8a1687cec8c8bd883ec73ca41d1",
        )
        timestamp = 1000
        for i in range(num_msgs):
            writer.write(conn, timestamp + i, serialized_msg)


def generate_frontend_bag(filename):
    print(f"Generating frontend test bag: {filename}")
    if os.path.exists(filename):
        os.remove(filename)

    with Writer(filename) as writer:
        # 1. rosgraph_msgs/Log
        conn_log = writer.add_connection(
            topic="/rosout",
            msgtype="rosgraph_msgs/msg/Log",
            msgdef="Header header\nbyte level\nstring name\nstring msg\nstring file\nstring function\nuint32 "
            "line\nstring[] topics\n================================================================================\n"
            "MSG: std_msgs/Header\nuint32 seq\ntime stamp\nstring frame_id",
            md5sum="acffd30cd6b6de30f120938c17c593fb",
        )
        # 2. sensor_msgs/Temperature
        conn_temp = writer.add_connection(
            topic="/sensors/temperature",
            msgtype="sensor_msgs/msg/Temperature",
            msgdef="Header header\nfloat64 temperature\nfloat64 variance"
            "\n================================================================================\n"
            "MSG: std_msgs/Header\nuint32 seq\ntime stamp\nstring frame_id",
            md5sum="ff71b307acdbe7c871a5a6d7edce2f6e",
        )
        # 3. sensor_msgs/TimeReference
        conn_time = writer.add_connection(
            topic="/time_ref",
            msgtype="sensor_msgs/msg/TimeReference",
            msgdef="Header header\ntime time_ref\nstring source\n"
            "================================================================================\n"
            "MSG: std_msgs/Header\nuint32 seq\ntime stamp\nstring frame_id",
            md5sum="fded64a0265108ba86c3d38fb11c0c16",
        )
        # 4. geometry_msgs/TwistStamped
        conn_twist = writer.add_connection(
            topic="/cmd_vel",
            msgtype="geometry_msgs/msg/TwistStamped",
            msgdef="Header header\ngeometry_msgs/Twist twist\n"
            "================================================================================\n"
            "MSG: std_msgs/Header\nuint32 seq\ntime stamp\nstring frame_id\n"
            "================================================================================\n"
            "MSG: geometry_msgs/Twist\nVector3 linear\nVector3 angular\n"
            "================================================================================\n"
            "MSG: geometry_msgs/Vector3\nfloat64 x\nfloat64 y\nfloat64 z",
            md5sum="98d34b0043a2093cf9d9345ab6eef12e",
        )
        # 5. tf2_msgs/TFMessage
        conn_tf = writer.add_connection(
            topic="/tf",
            msgtype="tf2_msgs/msg/TFMessage",
            msgdef="geometry_msgs/TransformStamped[] transforms\n"
            "================================================================================\n"
            "MSG: geometry_msgs/TransformStamped\nHeader header\nstring child_frame_id\n"
            "geometry_msgs/Transform transform\n"
            "================================================================================\n"
            "MSG: std_msgs/Header\nuint32 seq\ntime stamp\nstring frame_id\n"
            "================================================================================\n"
            "MSG: geometry_msgs/Transform\ngeometry_msgs/Vector3 translation\n"
            "geometry_msgs/Quaternion rotation\n"
            "================================================================================\n"
            "MSG: geometry_msgs/Vector3\nfloat64 x\nfloat64 y\nfloat64 z\n"
            "================================================================================\n"
            "MSG: geometry_msgs/Quaternion\nfloat64 x\nfloat64 y\nfloat64 z\nfloat64 w",
            md5sum="94810edda583a504dfda3829e70d7eec",
        )

        # Write messages
        for i in range(100):
            timestamp = 1000 + i * 100000000  # 100ms steps
            secs = int(timestamp / 1000000000)
            nsecs = timestamp % 1000000000

            # Log
            writer.write(
                conn_log,
                timestamp,
                serialize_log(
                    i,
                    secs,
                    nsecs,
                    "",
                    2,
                    "test_node",
                    f"Log message {i}",
                    "test.cpp",
                    "main",
                    i,
                    ["/rosout"],
                ),
            )

            # Temperature (sine wave)
            import math

            temp = 20.0 + 5.0 * math.sin(i * 0.1)
            writer.write(
                conn_temp,
                timestamp,
                serialize_temperature(i, secs, nsecs, "sensor_frame", temp, 0.1),
            )

            # TimeReference
            writer.write(
                conn_time,
                timestamp,
                serialize_time_reference(i, secs, nsecs, "time_frame", secs, nsecs, "GPS"),
            )

            # TwistStamped (circle)
            writer.write(
                conn_twist,
                timestamp,
                serialize_twist_stamped(i, secs, nsecs, "base_link", [1.0, 0.0, 0.0], [0.0, 0.0, 0.5]),
            )

            # TF
            writer.write(
                conn_tf,
                timestamp,
                serialize_tf_message(
                    [
                        {
                            "seq": i,
                            "secs": secs,
                            "nsecs": nsecs,
                            "frame_id": "map",
                            "child_frame_id": "base_link",
                            "tx": i * 0.1,
                            "ty": 0.0,
                            "tz": 0.0,
                            "rx": 0.0,
                            "ry": 0.0,
                            "rz": 0.0,
                            "rw": 1.0,
                        }
                    ]
                ),
            )


def main():
    data_dir = os.path.join(os.path.dirname(__file__), "data")
    os.makedirs(data_dir, exist_ok=True)

    files = {
        "10_KB.bag": 10 * 1024,
        "50_KB.bag": 50 * 1024,
        "1_MB.bag": 1 * 1024 * 1024,
        "17_MB.bag": 17 * 1024 * 1024,
        "125_MB.bag": 125 * 1024 * 1024,
    }

    for filename, size in files.items():
        filepath = os.path.join(data_dir, filename)
        generate_bag(filepath, size)

    # Generate backend fixtures
    backend_fixtures_dir = os.path.join(os.path.dirname(__file__), "../../backend/tests/fixtures")
    os.makedirs(backend_fixtures_dir, exist_ok=True)
    generate_bag(os.path.join(backend_fixtures_dir, "test.bag"), 10 * 1024)
    generate_bag(os.path.join(backend_fixtures_dir, "to_delete.bag"), 10 * 1024)
    generate_bag(os.path.join(backend_fixtures_dir, "file1.bag"), 10 * 1024)
    generate_bag(os.path.join(backend_fixtures_dir, "file2.bag"), 10 * 1024)
    generate_bag(os.path.join(backend_fixtures_dir, "move_me.bag"), 10 * 1024)
    generate_bag(os.path.join(backend_fixtures_dir, "state_test.bag"), 10 * 1024)

    # Generate backend dummy MCAP and YAML
    with open(os.path.join(backend_fixtures_dir, "config.yaml"), "w") as f:
        f.write("test: true\nvalue: 123\n")
    with open(os.path.join(backend_fixtures_dir, "config.yml"), "w") as f:
        f.write("test: true\nvalue: 123\n")
    with open(os.path.join(backend_fixtures_dir, "test.mcap"), "wb") as f:
        f.write(b"\x89MCAP\x30\r\n")

    generate_frontend_bag(os.path.join(data_dir, "frontend_test.bag"))

    # Generate dummy MCAP and YAML
    with open(os.path.join(data_dir, "test.yaml"), "w") as f:
        f.write("test: true\nvalue: 123\n")

    with open(os.path.join(data_dir, "test.mcap"), "wb") as f:
        f.write(b"\x89MCAP\x30\r\n")  # Minimal magic bytes

    print("Done.")


if __name__ == "__main__":
    main()
